/*
   ----------------------------------------------------------------------------------------------------
   PropaDuino thermostatgesteuerte Mini-Gewächshaus-Heizung
   (p) 2020 Mr.Chili

   Dieser Code ist kostenlos und darf von jedermann verändert
   werden.


   ==== Disclaimer ====

   Diese Schaltung erzeugt Wärme für ein Mini-Gewächshaus mithilfe
   einer elektrischen Heizung.

   Bitte jederzeit an die Risiken und die Feuergefahr einer Heizschaltung
   denken. Beim Bau der Schaltung auf äußerste Sorgfalt achten. Für
   Schäden die durch unsachgemäßen Nachbau und Betrieb entstehen wird
   keine Haftung übernommen.

   Wenn die Schaltung fertig ist, bitte im Betrieb einige Zeit im 
   Auge behalten und auf unbeabsichtigte Schaltungsfehler bzw.
   Fehlfunktionen achten, bevor die Heizung unbeaufsichtigt angelassen wird.

   

   ----------------------------------------------------------------------------------------------------
*/

// Libraries and pin definitions
#include <OneWire.h>
#include <DallasTemperature.h>

// Defining pins on the ATtiny85

// Heating pad pin for PWM
#define heater 0
// LED pin
#define LED 1
// OneWire pin for DS18B20 temperature sensor
#define ONE_WIRE_BUS 2
// Power mode sense pin
#define sensePin 4

// Defining low and high mode max temperatures.
// Adjust these to suit your needs.
// 23°C is fine for many plants.
// Nightshades like tomato or chili pepper
// like it a bit more cozy, 28°C is a good
// germination temperature for them.
#define lowModeTemp 23.00;
#define highModeTemp 28.00;

// Initializing the temperature sensor
OneWire oneWire(ONE_WIRE_BUS);
DallasTemperature sensors(&oneWire);

//--------- Global variables

float maxTemp;
int dutyCycle;
byte highMode;
int LEDdutyCycle;
byte cooldown = 0;

// Timekeeping variables
unsigned long systemStart = millis();
unsigned long checkTimestamp;
unsigned long timeTempCheckBegin = systemStart;

unsigned long heatingStart = 0;
unsigned long timeTempAverageBegin = systemStart;
unsigned long timeSensorCheckBegin = systemStart;

// 1- and 3-minute temperature check
float temperatureCheck1m;
float temperatureCheck3m;

// Blink routine variables
unsigned long blinkStart = systemStart;
unsigned long blinkStop;

// various counter variables
float tempCumul;
byte tempI = 0;
float tempAverage = 0;
byte quickheat = 0;
byte padFault = 0;

//--------- Blink function
// Controlling the status LED by telling it how to blink
// depending on the device status
void ledBlink(int blinkTime) {

  // check Time
  blinkStop = millis();

  // Blink LED on and off with the assigned time intervals
  if (blinkStop - blinkStart >= blinkTime) {

    if (bitRead(PORTB, PB1) == HIGH) bitWrite(PORTB, PB1, LOW);
    else bitWrite (PORTB, PB1, HIGH);
    blinkStart = blinkStop;
  }
}


void setup() {

  // Pin modes
  pinMode(heater, OUTPUT);
  pinMode(LED, OUTPUT);
  pinMode(sensePin, INPUT);
  sensors.begin();

  // Blink once to confirm start-up.
  // Blocking code does no harm here.
  digitalWrite(LED, HIGH);
  delay(500);
  digitalWrite(LED, LOW);
  delay(500);

  // Are we in high or low power mode?
  // Circuit is connected to a I-0-II rocker switch,
  // with pin 4 sensing if the switch is in the "I" position

  if (digitalRead(sensePin)) {

    //Setting low-mode max temperature.
    maxTemp = lowModeTemp;
    //Set PWM duty cycle for low mode
    dutyCycle = 130;
    highMode = 0;
    LEDdutyCycle = 30;
  }

  else {

    //Setting high-mode max temperature
    maxTemp = highModeTemp;
    //Set PWM duty cycle for high mode
    dutyCycle = 180;
    highMode = 1;
    LEDdutyCycle = 200;
  }

  // Getting starting temperature as reference
  // for first temperature and heating pad check
  sensors.requestTemperatures();

  float initialTemp = sensors.getTempCByIndex(0);

  // If temperature sensor isn't ready, set 20.00°C as the initial temperature.
  // This can prevent premature errors.
  if (initialTemp == -127.00) initialTemp = 20.00;

  tempAverage = initialTemp;

  temperatureCheck1m = initialTemp;
  temperatureCheck3m = initialTemp;
}

void loop() {

  // All-purpose timestamp on every cycle
  checkTimestamp = millis();

  // Resetting timekeeping on unsigned long rollover
  // after 49 days and 17 hours of continuous uptime
  if (checkTimestamp >= 4294966000) {

    checkTimestamp = 0;
    timeTempCheckBegin = 0;
    timeTempAverageBegin = 0;
    timeSensorCheckBegin = 0;

    blinkStart = 0;
  }


  // ------------ Main thermostat routine ---------------------------------------------------------------

  // "Lazy" temperature sensor averaging...
  // Average temperature every five seconds,
  // taking one reading a second from the DS18B20.

  // This can alternatively be done with an array of
  // constantly updated moving averages, but it will
  // not hugely improve functionality.
  if (tempI < 5 && (checkTimestamp - timeTempAverageBegin) >= 1000) {

    timeTempAverageBegin = checkTimestamp;
    sensors.requestTemperatures();
    float temperatureNow = sensors.getTempCByIndex(0);
    tempCumul += temperatureNow;
    tempI += 1;
  }

  if (tempI == 5) {

    // Calculating the arithmetic average of our five readings
    tempAverage = tempCumul / 5;
    tempCumul = 0;
    tempI = 0;
  }


  // If we get several readings of -127.00°C, the standard fault code
  // for the DS18B20, this must mean the temperature sensor is
  // not working properly. So if the average temperature from our
  // readings is calculated to be less than -50°C, let's shut the system
  // off and terminate in a blink loop.
  if (tempAverage < -50.00) {

    digitalWrite(heater, LOW);
    while (1) ledBlink(150);

  }

  // Quick heat-up on cold soil
  if (tempAverage <= 18.00 && tempAverage > -50.00) {

    analogWrite(heater, 220);
    quickheat = 1;
    analogWrite(LED, LEDdutyCycle);
  }

  // Or return to "regular" duty cycle after quick heat-up
  else  quickheat = 0;

  // Heat until max temperature is (almost) reached
  if (tempAverage < maxTemp && !cooldown && !quickheat && tempAverage > -50.00) {

    if (!heatingStart) heatingStart = checkTimestamp;
    analogWrite(heater, dutyCycle);
    analogWrite(LED, LEDdutyCycle);
  }

  // Shut off heat at max temperature
  if (tempAverage >= maxTemp) {

    cooldown = 1;
    digitalWrite(heater, LOW);
    digitalWrite(LED, LOW);

    heatingStart = 0;
  }

  // Allow time to cool off
  // Heater kicks back in when temperature goes down by more
  // than two degrees below maximum temperature
  if (tempAverage < (maxTemp - 2.00)) cooldown = 0;

  // Check heating progress every minute
  if (heatingStart && (checkTimestamp - heatingStart) >= 60000 && !quickheat) {

    // Adapt to changing ambient temperature and / or water content in the soil:

    // ...beef up heating pad duty cycle if heating is too slow
    if (tempAverage < (temperatureCheck1m + 0.25)  && !cooldown && tempAverage > -50.00) {

      if (highMode && dutyCycle <= 220) dutyCycle += 20;
      if (!highMode && dutyCycle <= 180) dutyCycle += 10;
    }

    // ...reduce heating pad duty cycle if heating is too fast
    if (tempAverage > (temperatureCheck1m + 0.75) && tempAverage > -50.00) {

      if (highMode && dutyCycle >= 140) dutyCycle -= 20;
      if (!highMode && dutyCycle >= 100) dutyCycle -= 10;
    }

    // Reset one-minute timer
    heatingStart = checkTimestamp;
    temperatureCheck1m = tempAverage;
  }

  // Checking heating pad every three minutes
  if ((checkTimestamp - timeTempCheckBegin) >= 180000) {

    if (heatingStart || quickheat) {

      // If the temperature goes down while heating and we're not in cooldown mode,
      // then there must be something wrong with the heating pad.
      if (tempAverage <= (temperatureCheck3m - 0.50) && !cooldown && tempAverage > -50) padFault = 1;
      else padFault = 0;

      timeTempCheckBegin = checkTimestamp;
      temperatureCheck3m = tempAverage;

    }
  }

  //If the heating pad is faulty, blink 250 ms
  if (padFault) ledBlink(250);
}
